'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _sparkCore = require('@ciscospark/spark-core');

/**
 * @typedef {Object} Types~Webhook
 * @property {string} id - The unique ID for the webhook.
 * @property {string} resource - The resource type for the webhook.
 * @property {string} event - The event type for the webhook.
 * @property {string} filter - The filter that defines the webhook scope.
 * @property {string} targetUrl - The URL that receives POST requests for each event.
 * @property {string} name - A user-friendly name for this webhook.
 */

/**
 * A webhook notifies an application when an event for which the application is
 * registered has occurred.
 * @class
 * @extends SparkPlugin
 */
var Webhooks = _sparkCore.SparkPlugin.extend({
  /**
   * Posts a webhook.
   * @instance
   * @memberof Webhooks
   * @param {Types~Webhook} webhook
   * @returns {Promise<Webhook>}
   * @example
   * var ciscospark = require('../..');
   * ciscospark.rooms.create({title: 'Create Webhook Example'})
   *   .then(function(room) {
   *     return ciscospark.webhooks.create({
   *       resource: 'messages',
   *       event: 'created',
   *       filter: 'roomId=' + room.id,
   *       targetUrl: 'https://example.com/webhook',
   *       name: 'Test Webhook'
   *     });
   *   })
   *   .then(function(webhook) {
   *     var assert = require('assert');
   *     assert(webhook.id);
   *     assert(webhook.resource);
   *     assert(webhook.event);
   *     assert(webhook.filter);
   *     assert(webhook.targetUrl);
   *     assert(webhook.name);
   *     return 'success';
   *   });
   *   // => success
   */
  create: function create(webhook) {
    return this.request({
      method: 'POST',
      uri: this.config.hydraServiceUrl + '/webhooks',
      body: webhook
    }).then(function (res) {
      return res.body;
    });
  },


  /**
   * Shows details for a webhook.
   * @instance
   * @memberof Webhooks
   * @param {Webhook|string} webhook
   * @returns {Promise<Array<Webhook>>}
   * @example
   * var ciscospark = require('../..');
   * var webhook;
   * ciscospark.rooms.create({title: 'Get Webhook Example'})
   *   .then(function(room) {
   *     return ciscospark.webhooks.create({
   *       resource: 'messages',
   *       event: 'created',
   *       filter: 'roomId=' + room.id,
   *       targetUrl: 'https://example.com/webhook',
   *       name: 'Test Webhook'
   *     });
   *   })
   *   .then(function(w) {
   *     webhook = w;
   *     return ciscospark.webhooks.get(webhook.id);
   *   })
   *   .then(function(webhook2) {
   *     var assert = require('assert');
   *     assert.deepEqual(webhook2, webhook);
   *     return 'success';
   *   });
   *   // => success
   */
  get: function get(webhook) {
    var id = webhook.id || webhook;

    return this.request({
      uri: this.config.hydraServiceUrl + '/webhooks/' + id
    }).then(function (res) {
      return res.body.items || res.body;
    });
  },


  /**
   * Lists all webhooks.
   * @instance
   * @memberof Webhooks
   * @param {Object} options
   * @param {integer} options.max Limit the maximum number of webhooks in the response.
   * @returns {Promise<Webhook>}
   * @example
   * var ciscospark = require('../..');
   * var room, webhook;
   * ciscospark.rooms.create({title: 'List Webhooks Example'})
   *   .then(function(r) {
   *     room = r;
   *     return ciscospark.webhooks.create({
   *       resource: 'messages',
   *       event: 'created',
   *       filter: 'roomId=' + room.id,
   *       targetUrl: 'https://example.com/webhook',
   *       name: 'Test Webhook'
   *     });
   *   })
   *   .then(function(w) {
   *     webhook = w;
   *     return ciscospark.webhooks.list();
   *   })
   *   .then(function(webhooks) {
   *     var assert = require('assert');
   *     assert.equal(webhooks.items.filter(function(w) {
   *       return w.id === webhook.id;
   *     }).length, 1);
   *     return 'success';
   *   });
   *   // => success
   */
  list: function list(options) {
    var _this = this;

    return this.request({
      uri: this.config.hydraServiceUrl + '/webhooks/',
      qs: options
    }).then(function (res) {
      return new _sparkCore.Page(res, _this.spark);
    });
  },


  /**
   * Delete a webhook.
   * @instance
   * @memberof Webhooks
   * @param {Webhook|string} webhook
   * @returns {Promise}
   * @example
   * var ciscospark = require('../..');
   * var room, webhook;
   * ciscospark.rooms.create({title: 'Remove Webhook Example'})
   *   .then(function(r) {
   *     room = r;
   *     return ciscospark.webhooks.create({
   *       resource: 'messages',
   *       event: 'created',
   *       filter: 'roomId=' + room.id,
   *       targetUrl: 'https://example.com/webhook',
   *       name: 'Test Webhook'
   *     });
   *   })
   *   .then(function(w) {
   *     webhook = w;
   *     return ciscospark.webhooks.remove(webhook);
   *   })
   *   .then(function() {
   *     return ciscospark.webhooks.list();
   *   })
   *   .then(function(webhooks) {
   *     var assert = require('assert');
   *     assert.equal(webhooks.items.filter(function(w) {
   *       return w.id === webhook.id;
   *     }).length, 0);
   *     return 'success';
   *   });
   *   // => success
   */
  remove: function remove(webhook) {
    var id = webhook.id || webhook;
    return this.request({
      method: 'DELETE',
      uri: this.config.hydraServiceUrl + '/webhooks/' + id
    }).then(function (res) {
      // Firefox has some issues with 204s and/or DELETE. This should move to
      // http-core
      if (res.statusCode === 204) {
        return undefined;
      }
      return res.body;
    });
  },


  /**
   * Update a webhook.
   * @instance
   * @memberof Webhooks
   * @param {Webhook} webhook
   * @returns {Promise<Webhook>}
   * @example
   * var ciscospark = require('../..');
   * var webhook;
   * ciscospark.rooms.create({title: 'Webhook Example'})
   *   .then(function(room) {
   *     return ciscospark.webhooks.create({
   *       resource: 'messages',
   *       event: 'created',
   *       filter: 'roomId=' + room.id,
   *       targetUrl: 'https://example.com/webhook',
   *       name: 'Test Webhook'
   *     });
   *   })
   *   .then(function(w) {
   *     webhook = w;
   *     webhook.targetUrl = 'https://example.com/webhook/newtarget';
   *     return ciscospark.webhooks.update(webhook);
   *   })
   *   .then(function() {
   *     return ciscospark.webhooks.get(webhook);
   *   })
   *   .then(function(webhook) {
   *     var assert = require('assert');
   *     assert.equal(webhook.targetUrl, 'https://example.com/webhook/newtarget');
   *     return 'success';
   *   });
   *   // => success
   */
  update: function update(webhook) {
    var id = webhook.id;
    return this.request({
      method: 'PUT',
      uri: this.config.hydraServiceUrl + '/webhooks/' + id,
      body: webhook
    }).then(function (res) {
      return res.body;
    });
  }
}); /**!
     *
     * Copyright (c) 2015-2017 Cisco Systems, Inc. See LICENSE file.
     * @private
     */

exports.default = Webhooks;